/* irsender.cc       letzte Aenderung: 17.7.2012
 Version: 0.3

 Prozessor: ATmega8A mit 16MHz Quarz
 Schaltung: Port B0 zum testen mit Frequenzmesser oder Oszilloskop verbunden
                 B1 zum testen mit LED
                 B2 mit IR-LED (ueber Transistor)
                 B3 mit roter LED
                 B4 mit gruener LED
            Tastenmatrix
		 D2 (INT0) Matrix-Ausgaenge ueber Dioden
		 D3 (INT1) nicht benutzt
		 fuer 4*5-Matrix:
		 C0-C5 Matrix-Eingaenge, 6 Reihen, als Schutz ueber 330-Ohm-Widerstaende anschliessen
		 D4-D7 Matrix-Ausgaenge, 4 Spalten

Skizze unterste 2 Reihen:
        /                /                /                /
C1 ---./ ,---+   C1 ---./ ,---+   C1 ---./ ,---+   C1 ---./ ,---+
             |                |                |                |
        /    |           /    |           /    |           /    |
C0 ---./ .---+   C0 ---./ .---+   C0 ---./ .---+   C0 ---./ .---+
             |                |                |                |
             |                |                |                |
            D4               D5               D6               D7
            _|_              _|_              _|_              _|_
            / \              / \              / \              / \
           /___\            /___\            /___\            /___\
             |                |                |                |
             +----------------+----------------+----------------+---INT0

 Autor: Rolf Pfister
 Copyright: Freeware
 History:
 Datum     Vers Kommentar
 1.7.2012  0.0  Erstellung
 2.7.12    0.1  mit Tastenmatrix
 3.7.12    0.2  mit zusaetzlichen Formaten fuer TV und Radio
 15.7.12   0.3  Formate fuer Sony-DVD-Blueray-Player, Pioneer-Harddisk-Recorder und Epson-Beamer
                Learn-Modus vorbereitet

*/
//#define OHNE_LEARNMODUS //provisorisch

#include <avr/io.h>
#include <avr/sleep.h>
#include <avr/interrupt.h>
#include "ulong.h"

#define TAKT_BIT 0  //zur Ausgabe der Modulationsfrequenz fuer Testzwecke 
#define LED1_BIT 1  //erste Kontroll-LED (keine LED angeschlossen)
#define LED2_BIT 3  //zweite Kontroll-LED (rot)
#define LED3_BIT 4  //dritte Kontroll-LED (gruen)
#define IRLED_BIT 2 //Ausgabe des Sendesignals

#define INT0_BIT PD2
#define INT1_BIT PD3

//Format fuer Kaon-DigitalTV-Empfaenger:
#define F_TAKT 38  //Modulations-Frequenz in kHz
#define PULSBREITE 21 //Anzahl Takte pro Puls
#define NULLBITLUECKE 21 //Anzahl Takte fuer 0-Bit
#define EINSBITLUECKE 63 //Anzahl Takte fuer 1-Bit
#define STARTHI 336 //Anzahl Takte fuer Startbit high
#define STARTLO 158 //Anzahl Takte fuer Startbit low
#define STARTLOREP 84 //Anzahl Takte fuer Startbit low im Repeat-Block
#define ADRESSE 0x0410 //Adresse des Geraetes
#define COD_POWER 0x00 //Code fuer Power-Taste
#define COD_1     0x38 //Code fuer 1-Taste
#define COD_2     0xB8 //Code fuer 2-Taste
#define COD_3     0x78 //Code fuer 3-Taste
#define COD_4     0x02 //Code fuer 4-Taste
#define COD_5     0x82 //Code fuer 5-Taste
#define COD_6     0x42 //Code fuer 6-Taste
#define COD_7     0x22 //Code fuer 7-Taste
#define COD_8     0xA2 //Code fuer 8-Taste
#define COD_9     0x62 //Code fuer 9-Taste
#define COD_0     0xE2 //Code fuer 0-Taste
#define COD_NEXT  0x60 //Code fuer Sender+
#define COD_PREV  0xE0 //Code fuer Sender-
#define COD_VPLUS 0x40 //Code fuer VOL+ Taste
#define COD_VMINUS 0xC0 //Code fuer VOL- Taste
#define COD_MUTE  0x20 //Code fuer Mute-Taste
#define COD_INFO  0xD8 //Code fuer Info-Taste
#define COD_EXIT  0x08 //Code fuer Exit-Taste
#define COD_RECAL 0x30 //Code fuer Recall-Taste

//Eigene Codes: (muessen verschieden sein zu allen obigen Codes)
#define COD_TV    0x01  //Fernseher
#define COD_RADIO 0x03  //Radio
#define COD_SAT   0x04  //Satellitempfaenger oder Kabelmodem (=DefaultGeraet)
#define COD_MODUS 0x05  //zum Modus weiterschalten
#define COD_REC   0x06  //Video- oder Harddisk-Recorder
#define COD_DVD   0x07  //DVD- oder Blueray-Player
#define COD_BEAM  0x09  //Beamer
#define COD_LEARNED 0xF0  //gelernte Taste

//Format fuer Samsung-Fernseher:
#define SAM_F_TAKT 38  //Modulations-Frequenz in kHz
#define SAM_PULSBREITE 21 //Anzahl Takte pro Puls
#define SAM_STARTHI 173 //Anzahl Takte fuer Startbit high
#define SAM_STARTLO 173 //Anzahl Takte fuer Startbit low
#define SAM_ADRESSE 0xE0E0 //Adresse des Geraetes
#define SAM_COD_POWER 0x40 //Code fuer Power-Taste
#define SAM_COD_QUELL 0x80 //Code fuer Source-Taste

//Format fuer Sony-Radio:
#define SON_F_TAKT 40  //Modulations-Frequenz in kHz
#define SON_STARTHI 96 //Anzahl Takte fuer Startbit high
#define SON_STARTLO 20 //Anzahl Takte fuer Startbit low
#define SON_NULLBIT 24 //Anzahl Takte fuer 0-Bit //Bits werden durch Breite der Pulse definiert
#define SON_EINSBIT 48 //Anzahl Takte fuer 1-Bit //Bits werden durch Breite der Pulse definiert
#define SON_LUECKE  20 //Anzahl Takte fuer Luecke zwischen Bits
#define SON_COD_POWER 0xA //4-Bit-Code fuer Power-Taste
#define SON_ADRESSE 0x81 //8-Bit-Adresse des Geraetes
#define SON_COD_VPLUS 0x4 //4-Bit-Code fuer VOL+ Taste
#define SON_COD_VMINUS 0xC //4-Bit-Code fuer VOL- Taste

//Format fuer Pioneer-Harddisk-Recorder:
#define PIO_F_TAKT 40  //Modulations-Frequenz in kHz
#define PIO_STARTHI 340 //Anzahl Takte fuer Startbit high
#define PIO_STARTLO 160 //Anzahl Takte fuer Startbit low
#define PIO_PULSBREITE 22 //Anzahl Takte pro Puls
#define PIO_NULLBITLUECKE 21 //Anzahl Takte fuer 0-Bit
#define PIO_EINSBITLUECKE 63 //Anzahl Takte fuer 1-Bit
#define PIO_PAKABSTAND 25 //Abstand in Millisekunden zum 2.Paket
#define PIO_ADRESSE 0xD52A //Adresse des Geraetes
#define PIO_COD_POWER 0x34 //Code im 1.Paket fuer Power-Taste und Zahlen-Tasten
#define PIO_ADRESSE2 0xF50A //Adresse im 2.Paket fuer Power-Taste und Zahlen-Tasten
#define PIO_COD_POWER2 0x3D //Code im 2.Paket fuer Power-Taste
#define PIO_COD_1 0x85 //Code im 2.Paket fuer 1-Taste
#define PIO_COD_2 0x45 //Code im 2.Paket fuer 2-Taste
#define PIO_COD_3 0xC5
#define PIO_COD_4 0x25
#define PIO_COD_5 0xA5
#define PIO_COD_6 0x65
#define PIO_COD_7 0xE5
#define PIO_COD_8 0x15
#define PIO_COD_9 0x95
#define PIO_COD_0 0x05
#define PIO_ADRESSE3 0xD52A //Adresse im 2.Paket fuer Spezial-Tasten
#define PIO_COD_PLAY 0x44 //Code fuer Play-Taste, im 1.Paket und im 2.Paket
#define PIO_COD_PAUS 0xC4 //Code fuer Pause-Taste
#define PIO_COD_STOP 0x04 //Code fuer Stopp-Taste
#define PIO_COD_TIMREC 0x54 //Code fuer TimerRec-Taste

//Format fuer Sony-DVD-Blueray-Player:

//Format fuer Epson-Beamer:

/**** Takterzeugung ****/
static volatile uchar taktzaehler=0,sendeflag=0;
static volatile uchar int0_flag=0,int1_flag=0;

// Genaue Quarzfrequenz in Hertz:
#define F_CPU 16000000L

#define NormalerCompwert (((F_CPU/F_TAKT+1000)/2000)-1) //ab 0.5 aufrunden, sonst abrunden
#define SonyCompwert (((F_CPU/SON_F_TAKT+1000)/2000)-1) //ab 0.5 aufrunden, sonst abrunden
#define PionCompwert (((F_CPU/PIO_F_TAKT+1000)/2000)-1) //ab 0.5 aufrunden, sonst abrunden

ISR(SIG_OUTPUT_COMPARE1A)
{
 static uchar takt=0;
 if((takt^=1)!=0)
  {
   taktzaehler++;
   PORTB |= (1<<TAKT_BIT); //PB0 setzen
   if(sendeflag!=0) PORTB |= (1<<IRLED_BIT); //falls sendeflag aktiv: IR-LED einschalten
  }
 else
  {
   PORTB &= ~(1<<TAKT_BIT); //PB0 loeschen
   PORTB &= ~(1<<IRLED_BIT); //IR-LED ausschalten
  }
}

void timer1_init()
{
 TCCR1B = (1<<WGM12)|(1<<CS10); //CTC-OCR1A-Modus, Prescaler=1 (also Timer-Takt ist gleich Quarz-Takt)
 OCR1A = NormalerCompwert;
 TCNT1 = 0; //Startwert des Timers
 TIMSK = 1<<OCIE1A; //Timer1 Interrupts bei Vergleichswert
}
void timer1_halt()
{ 
 TCCR1B = 0;
}

//Bei einem unerwarteten Interrupt nichts machen:
void __attribute__ ((naked)) __vector_default(void) {__asm__ ("reti");}

ISR(INT0_vect)
{
 int0_flag=1;
}

ISR(INT1_vect)
{
 int1_flag=1;
}

/**** Globale Variablen ****/
#define MAXLEARNJ 8  //Maximale Anzahl Tastendruecke pro gelernte Taste
#define MAXLEARNF 12 //Maximale Anzahl lernbare Tasten
static uchar learn[MAXLEARNJ*MAXLEARNF]; //provisorisch: waere besser im EEPROM

static uchar geraeteflag=0,modus=0;
#define MAXMODUS 3
#define GERAET_RADIO 1
#define GERAET_RECORDER 2

void led_einschalten()
{
 if(modus==0) PORTB |= (1<<LED2_BIT); //rote LED einschalten
 else         PORTB |= (1<<LED3_BIT); //gruene LED einschalten
}

void led_ausschalten()
{
 if(modus==0) PORTB &= ~(1<<LED2_BIT); //rote LED aus
 else         PORTB &= ~(1<<LED3_BIT); //gruene LED aus
}

void ir_startbit()
{
 taktzaehler=0;
 sendeflag=1; //Startbit starten
 led_einschalten();
 while(taktzaehler < (STARTHI/2)) ;//erste Haelfte des Startbits abwarten
 taktzaehler=0;
 while(taktzaehler < (STARTHI/2)) ;//zweite Haelfte des Startbits abwarten
 taktzaehler=0;
 sendeflag=0; //Low-Teil des Startbits starten
 led_ausschalten();
 while(taktzaehler < STARTLO) ;//Low-Teil des Startbits abwarten
}

void ir_sam_startbit()
{
 taktzaehler=0;
 sendeflag=1; //Startbit starten
 led_einschalten();
 while(taktzaehler < SAM_STARTHI) ;//Startbits abwarten
 taktzaehler=0;
 sendeflag=0; //Low-Teil des Startbits starten
 led_ausschalten();
 while(taktzaehler < SAM_STARTLO) ;//Low-Teil des Startbits abwarten
}

void ir_sony_startbit()
{
 taktzaehler=0;
 sendeflag=1; //Startbit starten
 led_einschalten();
 while(taktzaehler < SON_STARTHI) ;//Startbits abwarten
 taktzaehler=0;
 sendeflag=0; //Low-Teil des Startbits starten
 led_ausschalten();
 while(taktzaehler < SON_STARTLO) ;//Low-Teil des Startbits abwarten
}

void ir_bit_senden(uchar data)  //einzelnes Bit ueber IR-Diode senden
{
 uchar luecke;
 sendeflag = 1; //Puls fuer aktuelles Bit senden
 taktzaehler=0;
 led_einschalten();
 while(taktzaehler < PULSBREITE) ;//Puls fuer aktuelles Bit abwarten
 sendeflag = 0; //Luecke fuer aktuelles Bit senden
 taktzaehler=0;
 led_ausschalten();
 luecke = (data==0) ? NULLBITLUECKE : EINSBITLUECKE;
 while(taktzaehler < luecke) ;//Luecke fuer aktuelles Bit abwarten
}

void ir_byte_senden(uchar data)  //ein Byte ueber IR-Diode senden
{
 uchar maske;
 for(maske=0x80;maske!=0;maske>>=1)
  {
   ir_bit_senden(data&maske);
  }
}

void ir_repeat()
{
 //Hier sollte noch ein repeat-code gesendet werden,
 //ist aber fuer Kaon nicht unbedingt noetig.
 taktzaehler=0;
 sendeflag=1; //Startbit starten
 led_einschalten();
 while(taktzaehler < (STARTHI/2)) ;//erste Haelfte des Startbits abwarten
 taktzaehler=0;
 while(taktzaehler < (STARTHI/2)) ;//zweite Haelfte des Startbits abwarten
 taktzaehler=0;
 sendeflag=0; //Low-Teil des Startbits starten
 led_ausschalten();
 while(taktzaehler < STARTLOREP) ;//Low-Teil des Startbits abwarten
 ir_bit_senden(1);
}

void ir_sony_4bit_senden(uchar data)
{
 uchar i,n;
 for(i=0;i<4;i++,data<<=1)
  {
   sendeflag = 1; //Puls fuer aktuelles Bit senden
   taktzaehler=0;
   led_einschalten();
   n = (data&0x08)==0 ? SON_NULLBIT : SON_EINSBIT;
   while(taktzaehler < n) ;//Puls fuer aktuelles Bit abwarten
   sendeflag = 0; //Luecke fuer aktuelles Bit senden
   taktzaehler=0;
   led_ausschalten();
   while(taktzaehler < SON_LUECKE) ;//Luecke fuer aktuelles Bit abwarten
  }
}

void ir_sony_byte_senden(uchar data)
{
 ir_sony_4bit_senden(data>>4);
 ir_sony_4bit_senden(data);
}

void milliwait(int n)
{
 taktzaehler=0;
 while(n>0)
  {
   while(taktzaehler<F_TAKT) ;//eine Millisekunde abwarten
   taktzaehler=0;
   --n;
  }
}

static uchar tastencode[6*4] = 
 {
  COD_RADIO, COD_TV,      COD_POWER, COD_MUTE,  //Radio-Power TV-Power SAT-Power Stumm
  COD_MODUS, SAM_COD_QUELL, COD_REC, COD_VPLUS, //Modus       TV-Quell Recorder  VOL+
  COD_INFO, COD_EXIT,    COD_RECAL, COD_VMINUS, //Info        Exit     Recall    VOL-
  COD_7, COD_8, COD_9, COD_0,     //7 8 9 0
  COD_4, COD_5, COD_6, COD_NEXT,  //4 5 6 Sender+
  COD_1, COD_2, COD_3, COD_PREV   //1 2 3 Sender-
 };

static uchar tastencode2[6*4] = 
 {
  COD_BEAM,  COD_DVD,     COD_POWER, COD_MUTE,  //Radio-Power TV-Power SAT-Power Stumm
  COD_MODUS, SAM_COD_QUELL, COD_REC, COD_VPLUS, //Modus       TV-Quell Recorder  VOL+
  COD_INFO, COD_EXIT,    COD_RECAL, COD_VMINUS, //Info        Exit     Recall    VOL-
  COD_7, COD_8, COD_9, COD_0,     //7 8 9 0
  COD_4, COD_5, COD_6, COD_NEXT,  //4 5 6 Sender+
  COD_1, COD_2, COD_3, COD_PREV   //1 2 3 Sender-
 };

void led_blinken(uchar b)
{
 PORTB |= b; //LED einschalten
 milliwait(500);
 PORTB &= ~b; //LED aus
}

void auf_taste_reagieren(uchar n);
void learn_modus_behandeln(uchar n,uchar m);

int main(void)
{
 DDRB  = 0x1F; //PortB PB0-4 als Ausgaenge
 PORTB = 0; //alle Ausgaenge aus
 //DDRD = 0; //alles Eingaenge
 PORTD = (1<<PD2) | (1<<PD3) | (1<<PD4) | (1<<PD5) | (1<<PD6) | (1<<PD7); //Pullup-Widerstaende setzen
 DDRC = 0x3F; //PC0 bis PC5 als Ausgaenge
 PORTC = 0; //PC alle Ausgaenge auf LOW (=aktiv)

 //Interrupts vorbereiten:
 MCUCR |= (1<<ISC01);//INT0 bei fallender Flanke
 GICR |= (1<<INT0);//INT0-Interrupts erlauben
 // MCUCR |= (1<<ISC11);//INT1 bei fallender Flanke
 // GICR |= (1<<INT1);//INT1-Interrupts erlauben
 timer1_init();
 sei(); //Interrupt einschalten

 while(1) //Hauptschlaufe
  {
   if(int0_flag!=0)
    {
     //Entprellung:
     milliwait(10);
     if(int0_flag!=0 && (PIND & (1<<INT0_BIT))!=0) int0_flag=0; //wenn Taste schon wieder losgelassen --> Stoerpuls, ignorieren

     if(int0_flag!=0)
      {
       uchar i,j=0,m,n;
       for(i=0x01;i!=0x40;i<<=1) //Leitungen C0 bis C5 der Reihe nach auf Low
	{
	 PORTC = i^0xFF;
	 milliwait(1);
	 if((n=(PIND&0xF0))!=0xF0) //ist einer von PD4 bis PD7 auf low?
	  {
	   n>>=4;
	   n ^= 0x0F;
	   for(j=0x01;j!=n;j<<=1) ;//entsprechende Reihe suchen
	   break;
	  }
	}
       if(j!=0) //Reihe und Spalte der Taste gefunden
	{
	 for(m=0;j!=1;j>>=1)  m++; //Spalte zaehlen
	 for(;i!=1;i>>=1)    m+=4; //Reihe zaehlen
	 n = (modus!=0) ? tastencode2[m] : tastencode[m];
#ifndef OHNE_LEARNMODUS
	 if(modus>=2) //Behandlung des Learn-Modus
	  {
	   learn_modus_behandeln(n,m);
	  }
	 else
#endif
         if(n==COD_MODUS)
	  {
	   if(++modus==MAXMODUS) modus=0;
	   if(modus==0)      led_blinken(1<<LED2_BIT); //1.Modus: vorprogrammierte Tasten verwenden, rot blinken
	   else if(modus==1) led_blinken(1<<LED3_BIT); //2.Modus: gelernte Tasten verwenden, gruen blinken
	   else //if(modus==2) //Learn-Modus
	     led_blinken((1<<LED2_BIT)|(1<<LED3_BIT)); //rot+gruen = gelb blinken
	  }
#ifndef OHNE_LEARNMODUS
	 else if(n==COD_LEARNED && m<MAXLEARNF) //Gelernte Taste
	  {
	   uchar mj,j,maxj;
	   mj = m*MAXLEARNJ;
	   maxj = learn[mj++];
	   if(maxj<=MAXLEARNJ)
	    for(j=1;j<maxj;j++)
	     {
	      auf_taste_reagieren(learn[mj++]);
	      milliwait(200); //provisorisch: Wartezeit noch ausprobieren
	     }
	  }
#endif
	 else
	  {
	   auf_taste_reagieren(n);
	  }
	}
       else //Fehler: Reihe nicht gefunden
	{//test
	 for(uchar i=0;i<3;i++)  //test: LED 3 mal blinken lassen
	  {
	   PORTB |= (1<<LED2_BIT); //LED2 einschalten
	   milliwait(200);
	   PORTB &= ~(1<<LED2_BIT); //LED2 aus
	   milliwait(200);
	  }
	}
       milliwait(50);
       taktzaehler=0;
       int0_flag=0;
       PORTC = 0; //PC alle Ausgaenge auf LOW (=aktiv)
      }

    }
   if(int1_flag!=0)
    {
     for(uchar i=0;i<2;i++)  //test: LED 2 mal blinken lassen
	  {
	   PORTB |= (1<<LED2_BIT); //LED2 einschalten
	   milliwait(200);
	   PORTB &= ~(1<<LED2_BIT); //LED2 aus
	   milliwait(200);
	  }
     taktzaehler=0;
     int1_flag=0;
    }

   timer1_halt();
   MCUCR &= ~((1<<ISC01)|(1<<ISC00));//INT0 bei LOW //test
   //MCUCR &= ~((1<<ISC11)|(1<<ISC10));//INT1 bei LOW

   //moegliche Schlafmodi: SLEEP_MODE_IDLE, _PWR_DOWN, _PWR_SAVE, _ADC, _STANDBY, _EXT_STANDBY
   set_sleep_mode(SLEEP_MODE_PWR_DOWN); //Schlafmodus setzen
   sleep_mode(); //Schlafmodus wirklich einschalten

   //MCUCR |= (1<<ISC11);//INT1 bei fallender Flanke
   MCUCR |= (1<<ISC01);//INT0 bei fallender Flanke
   timer1_init();
  }

 return 0; //wird nie erreicht.
}

void auf_taste_reagieren(uchar n)
{
 if(n==COD_TV || n==SAM_COD_QUELL) //Samsung-Fernseher
	  {
	   uchar cod;
	   cod = n==COD_TV ? SAM_COD_POWER : n;
	   ir_sam_startbit();
	   ir_byte_senden(SAM_ADRESSE>>8);//Geraeteadresse senden
	   ir_byte_senden(SAM_ADRESSE&0xFF);//Geraeteadresse senden
	   ir_byte_senden(cod);//Code fuer Powertaste senden
	   ir_byte_senden(cod^0xFF);//Inverser Code fuer Powertaste senden
	   ir_bit_senden(0);//abschliessendes Bit senden
	   do {milliwait(200); ir_repeat();}
	   while((PIND & (1<<INT0_BIT))==0);
	   geraeteflag=0;
	  }
 else if(n==COD_RADIO || (geraeteflag==GERAET_RADIO && (n==COD_VPLUS || n==COD_VMINUS)) ) //Sony-Radio
	  {
	   uchar cod;
	   cod = (n==COD_RADIO) ? SON_COD_POWER : ((n==COD_VPLUS) ? SON_COD_VPLUS : SON_COD_VMINUS);
	   do
	    {
	     for(uchar j=0;j<3;j++) //fuer Sony-Format 3 mal senden mit 25ms Abstand
	      {
	       OCR1A=SonyCompwert;
	       ir_sony_startbit();
	       ir_sony_4bit_senden(cod);//Code fuer Powertaste oder Lautstaerkereglung senden
	       ir_sony_byte_senden(SON_ADRESSE);//Geraeteadresse senden
	       OCR1A=NormalerCompwert;
	       milliwait(25);
	      }
	     milliwait(500);
	    }
	   while((PIND & (1<<INT0_BIT))==0);
	   geraeteflag=GERAET_RADIO;
	  }
 else if(n==COD_REC) //Pioneer-Harddisk-Recorder
	  {
	   OCR1A=PionCompwert;
	   ir_startbit();
	   ir_byte_senden(PIO_ADRESSE>>8);//Geraeteadresse senden
	   ir_byte_senden(PIO_ADRESSE&0xFF);//Geraeteadresse senden
	   ir_byte_senden(PIO_COD_POWER);//Code fuer Powertaste senden
	   ir_byte_senden(PIO_COD_POWER^0xFF);//Inverser Code fuer Powertaste senden
	   ir_bit_senden(0);//abschliessendes Bit senden
	   OCR1A=NormalerCompwert;
	   milliwait(PIO_PAKABSTAND);
	   OCR1A=PionCompwert;
	   ir_startbit();
	   ir_byte_senden(PIO_ADRESSE2>>8);//Geraeteadresse senden
	   ir_byte_senden(PIO_ADRESSE2&0xFF);//Geraeteadresse senden
	   ir_byte_senden(PIO_COD_POWER2);//Code fuer Powertaste senden
	   ir_byte_senden(PIO_COD_POWER2^0xFF);//Inverser Code fuer Powertaste senden
	   ir_bit_senden(0);//abschliessendes Bit senden
	   OCR1A=NormalerCompwert;
	   milliwait(500);
	   geraeteflag=GERAET_RECORDER;
	  }
 else if(n==COD_DVD) //DVD- oder Blueray-Player
	  {
	  }
 else if(n==COD_BEAM) //Beamer
	  {
	  }
 //else if(n==COD_SAT) //Default-Geraet
 //	  {
 //	  }
 else //alle andern Codes fuer Default-Geraet
	  {
	   ir_startbit();
	   ir_byte_senden(ADRESSE>>8);//Geraeteadresse senden
	   ir_byte_senden(ADRESSE&0xFF);//Geraeteadresse senden
	   ir_byte_senden(n);//Code fuer entsprechende Taste senden
	   ir_byte_senden(n^0xFF);//Inverser Code fuer Powertaste senden
	   ir_bit_senden(0);//abschliessendes Bit senden
	   do {milliwait(200); ir_repeat();}
	   while((PIND & (1<<INT0_BIT))==0);
	   geraeteflag=0;
	  }
}

void learn_modus_behandeln(uchar n,uchar m)
{
 static uchar learntaste=0,learnj=0;
 if(n==COD_MODUS || (modus==2 && m>=MAXLEARNF)) //erneutes Druecken der Modus-Taste? oder falsche Taste?
	    {
	     if(modus==2) //ohne Learn-Sequenz nochmals gedrueckt? oder falsche Taste gedrueckt?
	      {
	       modus=0; //ja: dann wieder auf Normalmodus wechseln
	       led_blinken(1<<LED2_BIT); //rote LED blinken lassen
	      }
	     else if(modus==3) //Learn-Modus abschliessen
	      {
	       learn[learntaste*MAXLEARNJ]=learnj; //erster Wert im Learn-Speicher auf Anzahl setzen
	       tastencode2[learntaste]=COD_LEARNED;
	       modus=1; //auf 2.Modus wechseln (zum Verwenden der gelernten Tasten)
	       led_blinken(1<<LED3_BIT); //gruene LED blinken lassen
	      }
	    }
 else if(modus==2) //erste andere Taste gedrueckt?
	    {
	     learntaste=m; //ja: als zu lernende Taste speichern
	     learnj=1;
	     modus++;
	     led_blinken((1<<LED2_BIT)|(1<<LED3_BIT)); //rot+gruen = gelb blinken
	    }
 else if(learnj<MAXLEARNJ) //noch Platz im learn-Speicher?
	    {
	     learn[learntaste*MAXLEARNJ+learnj++]=n; //ja: Taste speichern
	     led_blinken((1<<LED2_BIT)|(1<<LED3_BIT)); //rot+gruen = gelb blinken
	    }
 else
	    {
	     milliwait(500); //bei vollem Speicher nur kurz warten und nicht blinken
	    }
}
